package LATMOS::Accounts::Bases::Sql::objects;

use 5.010000;
use strict;
use warnings;

use base qw(LATMOS::Accounts::Bases::Objects);
use LATMOS::Accounts::Log;

our $VERSION = (q$Rev$ =~ /^Rev: (\d+) /)[0];

=head1 NAME

LATMOS::Ad - Perl extension for blah blah blah

=head1 SYNOPSIS

  use LATMOS::Accounts::Bases;
  my $base = LATMOS::Accounts::Bases->new('unix');
  ...

=head1 DESCRIPTION

Account base access over standard unix file format.

=cut

# This fields are special in sense they may come from site field
sub _address_fields { qw(l postalCode streetAddess postOfficeBox st) }

=head1 FUNCTIONS

=cut

=head2 new(%options)

Create a new LATMOS::Ad object for windows AD $domain.

domain / server: either the Ad domain or directly the server

ldap_args is an optionnal list of arguments to pass to L<Net::LDAP>.

=cut

sub list {
    my ($class, $base) = @_;

    my $sth = $base->db->prepare_cached(
        sprintf(
            q{select %s as k from %s %s order by %s},
            $base->db->quote_identifier($class->key_field),
            $base->db->quote_identifier($class->object_table),
            ($base->{wexported} ? '' : 'where exported = true'),
            $base->db->quote_identifier($class->key_field),
        )
    );
    $sth->execute;
    my @keys;
    while(my $res = $sth->fetchrow_hashref) {
        push(@keys, $res->{k});
    }
    @keys
}

sub list_from_rev {
    my ($class, $base, $rev) = @_;
    my $sth = $base->db->prepare_cached(
        sprintf(
            q{select %s as k from %s where rev > ? %s order by %s},
            $base->db->quote_identifier($class->key_field),
            $base->db->quote_identifier($class->object_table),
            ($base->{wexported} ? '' : 'and exported = true'),
            $base->db->quote_identifier($class->key_field),
        )
    );
    $sth->execute($rev);
    my @keys;
    while(my $res = $sth->fetchrow_hashref) {
        push(@keys, $res->{k});
    }
    @keys
}

sub has_extended_attributes { 0 }

# Work only for very simple case, must be override

sub _inline_fields {
    my ($self, $for, $base) = @_;
    my @res;
    my $sth = $base->db->prepare(
        q{SELECT column_name FROM information_schema.columns
          WHERE table_name = ?}
    );
    $sth->execute($self->object_table);
    while (my $res = $sth->fetchrow_hashref) {
        if ($for =~ 'w') {
            next if($res->{column_name} =~ /^(rev|date|create|ikey)$/);
            next if($res->{column_name} eq $self->key_field);
        }
        push(@res, $res->{column_name});
    }
    my %fields = map { $_ => $_ } @res;
    %fields
}

# Everything managed by the perl code

sub _managed_fields {
    my ($class, $for, $base) = @_;
    return();
}

# Everything from attributes_list table
# $for is uneeded here as all this attributes are rw

sub _extended_field {
    my ($class, $for, $base) = @_;
    my @attr;
    if ($class->has_extended_attributes) {
        my $sth = $base->db->prepare_cached(
            sprintf(
                q{select canonical from %s order by canonical},
                $base->db->quote_identifier($class->object_table . '_attributes_list'),
            )
        );
        $sth->execute;
        while (my $res = $sth->fetchrow_hashref) {
            push(@attr, $res->{canonical});
        }
        return map { $_ => $_ } @attr;
    } else {
        return ()
    }
}

sub _canonical_fields {
    my ($class, $base, $for) = @_;
    $for ||= 'rw';
    my %inl = (
        ($class->_inline_fields($for, $base)),
        ($class->_managed_fields($for, $base)),
        ($class->_extended_field($for, $base)),
    );
    return sort keys %inl;
}

sub _get_field_name_db {
    my ($class, $c_field, $base) = @_;
    $class->has_extended_attributes or return;
    $class->object_table or return;
    my $sth = $base->db->prepare_cached(
        sprintf(
            q{select ikey from %s where canonical = ?},
            $base->db->quote_identifier($class->object_table . '_attributes_list'),
        )
    );
    $sth->execute($c_field);
    
    my $res = $sth->fetchrow_hashref;
    $sth->finish;
    return $c_field if($res->{ikey});
}

sub _get_field_name {
    my ($class, $c_field, $base, $for) = @_;
    $c_field or return;
    $for ||= 'rw';
    my %fields = $class->_managed_fields($for, $base);
    return $fields{$c_field} if ($fields{$c_field});
    %fields = $class->_inline_fields($for, $base);
    return $fields{$c_field} if ($fields{$c_field});
    %fields = $class->_extended_field($for, $base);
    return $fields{$c_field}
}

sub new {
    my ($class, $base, $id) = @_;
    my $sth = $base->db->prepare_cached(
        sprintf(q{select 1 from %s where %s = ? %s},
            $base->db->quote_identifier($class->object_table),
            $base->db->quote_identifier($class->key_field),
            ($base->{wexported} ? '' : 'and exported = true'),
        ),
    );
    my $count = $sth->execute($id);
    $sth->finish;
    $count == 1 or return;
    bless({ _id => $id, _base => $base }, $class);
}

sub _get_ikey {
    my ($class, $base, $id) = @_;
    $base ||= $class->base;
    $id ||= $class->id;
    my $sth = $base->db->prepare_cached(
        sprintf(
            q{select ikey from %s where %s = ?},
            $base->db->quote_identifier($class->object_table),
            $base->db->quote_identifier($class->key_field),
        )
    );
    $sth->execute($id);
    my $res = $sth->fetchrow_hashref;
    $sth->finish;
    $res->{ikey}
}

sub _create {
    my ($class, $base, $id, %data) = @_;

    # splitting inline from extended
    my %inlined = $class->_inline_fields('w', $base);
    my %inl = map { $_ => 1 } values %inlined;
    my (%first, %second);
    foreach (keys %data) {
        warn $_;
        if ($inl{$_}) {
            $first{$_} = $data{$_};
        } else {
            $second{$_} = $data{$_};
        }
    }
    $first{$class->key_field} = $id;

    my $sth = $base->db->prepare(
        sprintf(
            q{insert into %s (%s) values (%s)},
            $base->db->quote_identifier($class->object_table),
            join(', ', map { $base->db->quote_identifier($_) } sort keys %first),
            join(',', qw(?) x scalar(keys %first)),
        )
    );
    $sth->execute(map { $first{$_} } sort keys %first) or return;

    $class->new($base, $id)->set_fields(%second);

    1;
}

sub _delete {
    my ($class, $base, $id) = @_;

    my $sthd = $base->db->prepare_cached(
        sprintf(
            q{delete from %s where %s = ?},
            $base->db->quote_identifier($class->object_table),
            $base->db->quote_identifier($class->key_field),
        )
    );
    $sthd->execute($id);
}

sub db {
    return $_[0]->base->db;
}

sub quote_object_table {
    my ($self) = @_;
    my $table = $self->object_table or return;
    $self->db->quote_identifier($table);
}
sub quote_key_field { 
    my ($self) = @_;
    my $key_field = $self->key_field or return;
    $self->db->quote_identifier($key_field);
}

sub get_field {
    my ($self, $field) = @_;
    my %inl = $self->_inline_fields('r', $self->base);
    my %inline = map { $_ => 1 } values %inl;
    if ($inline{$field}) {
    my $sth = $self->db->prepare_cached(
        sprintf(
            q{select %s from %s where %s = ?},
            $self->db->quote_identifier(lc($field)),
            $self->quote_object_table,
            $self->quote_key_field,
        )
    );
    $sth->execute($self->id);
    my $res = $sth->fetchrow_hashref;
    $sth->finish;
    return $res->{$field};
    } elsif ($self->has_extended_attributes) { # else, then we mandatory have extend attr
        my $sth = $self->db->prepare_cached(
            sprintf(
                q{
                select value from %s 
                join %s on okey = ikey
                where %s = ? and attr = ?
                },
                $self->db->quote_identifier($self->object_table. '_attributes'),
                $self->db->quote_identifier($self->object_table),
                $self->db->quote_identifier($self->key_field),
            )
        );
        $sth->execute($self->id, $field);
        my $res = $sth->fetchrow_hashref;
        $sth->finish;
        return $res->{value};
    }
}

sub set_fields {
    my ($self, %data) = @_;
    my @fields;
    my @vals;
    my %ext;
    my %inl = $self->_inline_fields('w', $self->base);
    my %inline = map { $_ => 1 } values %inl;
    foreach my $field (keys %data) {
        $data{$field} = $data{$field} ? 1 : 0 if($field eq 'exported');
        my $oldval = $self->get_field($field);
        next if (($data{$field} || '') eq ($oldval || ''));
        if ($inline{$field}) {
        # TODO check fields exists !
            push(@fields, sprintf("%s = ?", $self->db->quote_identifier($field)));
            push(@vals, $data{$field});
        } else {
            $ext{$field} = $data{$field};
        }
    }
    if (@fields) {
        my $sth = $self->db->prepare_cached(
            sprintf(
                q{update %s set %s where %s = ?},
                $self->quote_object_table,
                join(', ', @fields),
                $self->quote_key_field,
            )
        );
        $sth->execute(@vals, $self->id) or do {
            $self->base->log(LA_ERR, "Cannot update inline field %s" .
                $self->base->db->strerr);
            return;
        };
    }
    
    if ($self->has_extended_attributes) {
        my $sthd = $self->db->prepare_cached(
            sprintf(
                q{delete from %s where okey = ? and attr = ?},
                $self->db->quote_identifier($self->object_table. '_attributes'),
            ),
        );
        my $sthx = $self->db->prepare_cached(
            sprintf(
                q{insert into %s (okey, attr, value) values (?,?,?)},
                $self->db->quote_identifier($self->object_table. '_attributes'),
            )
        );
        my $sthu = $self->db->prepare_cached(
            sprintf(
                q{update %s set value = ? where okey = ? and attr = ?},
                $self->db->quote_identifier($self->object_table. '_attributes'),
            )
        );

        my $okey = $self->_get_ikey($self->base, $self->id);
        foreach (keys %ext) {
            if ($ext{$_}) {
                my $res = $sthu->execute($ext{$_}, $okey, $_);
                defined($res) or do {
                    $self->base->log(LA_ERR,
                        "Error while udapting attributes: %s",
                        $self->base->db->strerr
                    );
                    return;
                };
                if ($res == 0) {
                    $res = $sthx->execute($okey, $_, $ext{$_});
                    defined($res) or do {
                        $self->base->log(LA_ERR,
                            "Error while udapting attributes: %s",
                            $self->base->db->strerr
                        );
                        return;
                    };
                }
            } else {
                defined($sthd->execute($okey, $_)) or do {
                    $self->base->log(LA_ERR,
                        "Error while deleting attributes: %s",
                        $self->base->db->strerr
                    );
                    return;
                };
            }
        }
    }

    1;
}

sub attributes_summary {
    my ($class, $base, $attribute) = @_;
    $class->has_extended_attributes && $class->object_table or
        return $class->SUPER::attributes_summary($base, $attribute);
    my $sth = $base->db->prepare_cached(
        sprintf(
            q{select value from %s where attr = ? group by value},
            $base->db->quote_identifier($class->object_table .
                '_attributes'),
        )
    );
    $sth->execute($attribute);

    my @values;
    while (my $res = $sth->fetchrow_hashref) {
        push(@values, $res->{value});
    }
    @values
}


sub search {
    my ($class, $base, @filter) = @_;

    if ($class->has_extended_attributes) {
        my @sqlintersec;
        my @bind;
        while (my $item = shift(@filter)) {
            # attr=foo => no extra white space !
            # \W is false, it is possible to have two char
            my ($attr, $mode, $val) = $item =~ /^(\w+)(?:(\W)(.+))?$/ or next;
            if (!$mode) {
                $mode = '~';
                $val = shift(@filter);
            }
            push(@sqlintersec,
                sprintf(q{select okey from %s where attr=? %s},
                    $base->db->quote_identifier($class->object_table .
                        '_attributes'),
                    $val eq '*'
                        ? ''
                        : ($mode eq '~'
                            ? q{and value ILIKE ?}
                            : q{and value = ?} )
                )
            );
            push(@bind, $base->get_field_name($class->type, $attr, 'r'));
            push(@bind, $val) unless($val eq '*');
        }
        my $sth = $base->db->prepare(
            sprintf(q{
                select name from %s where ikey in (
                %s
                )
                },
                $base->db->quote_identifier($class->object_table),
                join("\n intersect\n", @sqlintersec),
            )
        );
        $sth->execute(@bind);
        my @results;
        while (my $res = $sth->fetchrow_hashref) {
            push(@results, $res->{name});
        }
        return(@results);
    } else {
        my @bind;
        my @where;
        while (my $item = shift(@filter)) {
            # attr=foo => no extra white space !
            # \W is false, it is possible to have two char
            my ($attr, $mode, $val) = $item =~ /^(\w+)(?:(\W)(.+))?$/ or next;
            if (!$mode) {
                $mode = '~';
                $val = shift(@filter);
            }
            if ($val eq '*') {
                push(@where, sprintf("%s is not NULL",
                        $base->db->quote_identifier($base->get_field_name($class->type,
                                $attr, 'r'))
                    )
                );
            } else {
                push(@where, sprintf("%s %s ?",
                        $base->db->quote_identifier(
                            $base->get_field_name($class->type, $attr, 'r')
                        ),
                        $mode eq '~' ? 'ILIKE' : '=',
                    ));
                push(@bind, lc($val));
            }
        }
        my $sth = $base->db->prepare(
            sprintf(q{select name from %s where %s},
                $base->db->quote_identifier($class->object_table),
                join(' and ', @where),
            )
        );
        $sth->execute(@bind);
        my @results;
        while (my $res = $sth->fetchrow_hashref) {
            push(@results, $res->{name});
        }
        return(@results);
    }
}

1;

__END__

=head1 SEE ALSO

=head1 AUTHOR

Olivier Thauvin, E<lt>olivier.thauvin@latmos.ipsl.frE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2008, 2009 CNRS SA/CETP/LATMOS

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.0 or,
at your option, any later version of Perl 5 you may have available.


=cut
